@@ -11,6 +11,8 @@
 #include <mruby/dump.h>
 #include <mruby/variable.h>
 #include <mruby/proc.h>
+#include <mruby/error.h>
+#include <mruby/presym.h>
 
 #if defined(_WIN32) || defined(_WIN64)
 # include <io.h> /* for setmode */
@@ -63,6 +65,16 @@ usage(const char *name)
     printf("  %s\n", *p++);
 }
 
+static mrb_bool
+mrb_extension_p(const char *path)
+{
+  const char *e = strrchr(path, '.');
+  if (e && e[1] == 'm' && e[2] == 'r' && e[3] == 'b' && e[4] == '\0') {
+    return TRUE;
+  }
+  return FALSE;
+}
+
 static void
 options_init(struct options *opts, int argc, char **argv)
 {
@@ -92,7 +104,7 @@ options_opt(struct options *opts)
     if (opts->opt[1] == '-') {
       /* `--` */
       if (!opts->opt[2]) {
-        ++opts->argv, --opts->argc;
+        opts->argv++, opts->argc--;
         return NULL;
       }
       /* long option */
@@ -102,7 +114,7 @@ options_opt(struct options *opts)
     }
     else {
       /* short option */
-      ++opts->opt;
+      opts->opt++;
       goto short_opt;
     }
   }
@@ -272,7 +284,6 @@ main(int argc, char **argv)
   mrb_value ARGV;
   mrbc_context *c;
   mrb_value v;
-  mrb_sym zero_sym;
 
   if (mrb == NULL) {
     fprintf(stderr, "%s: Invalid mrb_state, exiting mruby\n", *argv);
@@ -304,17 +315,14 @@ main(int argc, char **argv)
       c->no_exec = TRUE;
 
     /* Set $0 */
-    zero_sym = mrb_intern_lit(mrb, "$0");
+    const char *cmdline;
     if (args.rfp) {
-      const char *cmdline;
       cmdline = args.cmdline ? args.cmdline : "-";
-      mrbc_filename(mrb, c, cmdline);
-      mrb_gv_set(mrb, zero_sym, mrb_str_new_cstr(mrb, cmdline));
     }
     else {
-      mrbc_filename(mrb, c, "-e");
-      mrb_gv_set(mrb, zero_sym, mrb_str_new_lit(mrb, "-e"));
+      cmdline = "-e";
     }
+    mrb_gv_set(mrb, mrb_intern_lit(mrb, "$0"), mrb_str_new_cstr(mrb, cmdline));
 
     /* Load libraries */
     for (i = 0; i < args.libc; i++) {
@@ -326,7 +334,8 @@ main(int argc, char **argv)
         cleanup(mrb, &args);
         return EXIT_FAILURE;
       }
-      if (args.mrbfile) {
+      mrbc_filename(mrb, c, args.libv[i]);
+      if (mrb_extension_p(args.libv[i])) {
         v = mrb_load_irep_file_cxt(mrb, lfp, c);
       }
       else {
@@ -335,12 +344,15 @@ main(int argc, char **argv)
       fclose(lfp);
       e = mrb_vm_ci_env(mrb->c->cibase);
       mrb_vm_ci_env_set(mrb->c->cibase, NULL);
-      mrb_env_unshare(mrb, e);
+      mrb_env_unshare(mrb, e, FALSE);
       mrbc_cleanup_local_variables(mrb, c);
     }
 
+    /* set program file name */
+    mrbc_filename(mrb, c, cmdline);
+
     /* Load program */
-    if (args.mrbfile) {
+    if (args.mrbfile || mrb_extension_p(cmdline)) {
       v = mrb_load_irep_file_cxt(mrb, args.rfp, c);
     }
     else if (args.rfp) {
@@ -356,6 +368,7 @@ main(int argc, char **argv)
     mrb_gc_arena_restore(mrb, ai);
     mrbc_context_free(mrb, c);
     if (mrb->exc) {
+      MRB_EXC_CHECK_EXIT(mrb, mrb->exc);
       if (!mrb_undef_p(v)) {
         mrb_print_error(mrb);
       }
