@@ -36,7 +36,7 @@
 #       # foo: two
 #       # foo: three
 #
-# This allows you to chain Enumerators together.  For example, you
+# This allows you to chain Enumerators together. For example, you
 # can map a list's elements to strings containing the index
 # and the element as a string via:
 #
@@ -89,8 +89,7 @@ class Enumerator
   include Enumerable
 
   ##
-  # @overload initialize(size = nil, &block)
-  # @overload initialize(obj, method = :each, *args)
+  # @overload initialize(obj, method = :each, *args, **kwd)
   #
   # Creates a new Enumerator object, which can be used as an
   # Enumerable.
@@ -109,27 +108,32 @@ class Enumerator
   #
   #     p fib.take(10) # => [1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
   #
-  def initialize(obj=nil, meth=:each, *args, &block)
+  # In the second, deprecated, form, a generated Enumerator iterates over the
+  # given object using the given method with the given arguments passed. This
+  # form is left only for internal use.
+  #
+  # Use of this form is discouraged. Use Kernel#enum_for or Kernel#to_enum
+  # instead.
+  def initialize(obj=NONE, meth=:each, *args, **kwd, &block)
     if block
       obj = Generator.new(&block)
-    else
-      raise ArgumentError unless obj
-    end
-    if @obj and !self.respond_to?(meth)
-      raise NoMethodError, "undefined method #{meth}"
+    elsif obj == NONE
+      raise ArgumentError, "wrong number of arguments (given 0, expected 1+)"
     end
 
     @obj = obj
     @meth = meth
-    @args = args.dup
+    @args = args
+    @kwd = kwd
     @fib = nil
     @dst = nil
     @lookahead = nil
     @feedvalue = nil
     @stop_exc = false
   end
-  attr_accessor :obj, :meth, :args, :fib
-  private :obj, :meth, :args, :fib
+
+  attr_accessor :obj, :meth, :args, :kwd
+  attr_reader :fib
 
   def initialize_copy(obj)
     raise TypeError, "can't copy type #{obj.class}" unless obj.kind_of? Enumerator
@@ -137,6 +141,7 @@ class Enumerator
     @obj = obj.obj
     @meth = obj.meth
     @args = obj.args
+    @kwd = obj.kwd
     @fib = nil
     @lookahead = nil
     @feedvalue = nil
@@ -149,7 +154,7 @@ class Enumerator
   #   e.with_index(offset = 0)
   #
   # Iterates the given block for each element with an index, which
-  # starts from +offset+.  If no block is given, returns a new Enumerator
+  # starts from +offset+. If no block is given, returns a new Enumerator
   # that includes the index, starting from +offset+
   #
   # +offset+:: the starting index to use
@@ -221,13 +226,11 @@ class Enumerator
   end
 
   def inspect
-    return "#<#{self.class}: uninitialized>" unless @obj
-
     if @args && @args.size > 0
       args = @args.join(", ")
-      "#<#{self.class}: #{@obj}:#{@meth}(#{args})>"
+      "#<#{self.class}: #{@obj.inspect}:#{@meth}(#{args})>"
     else
-      "#<#{self.class}: #{@obj}:#{@meth}>"
+      "#<#{self.class}: #{@obj.inspect}:#{@meth}>"
     end
   end
 
@@ -243,9 +246,10 @@ class Enumerator
   #
   # === Examples
   #
-  #   "Hello, world!".scan(/\w+/)                     #=> ["Hello", "world"]
-  #   "Hello, world!".to_enum(:scan, /\w+/).to_a      #=> ["Hello", "world"]
-  #   "Hello, world!".to_enum(:scan).each(/\w+/).to_a #=> ["Hello", "world"]
+  #   Array.new(3)                     #=> [nil, nil, nil]
+  #   Array.new(3) { |i| i }           #=> [0, 1, 2]
+  #   Array.to_enum(:new, 3).to_a      #=> [0, 1, 2]
+  #   Array.to_enum(:new).each(3).to_a #=> [0, 1, 2]
   #
   #   obj = Object.new
   #
@@ -284,7 +288,7 @@ class Enumerator
   end
 
   def enumerator_block_call(&block)
-    @obj.__send__ @meth, *@args, &block
+    @obj.__send__ @meth, *@args, **@kwd, &block
   end
   private :enumerator_block_call
 
@@ -293,7 +297,7 @@ class Enumerator
   #   e.next   -> object
   #
   # Returns the next object in the enumerator, and move the internal position
-  # forward.  When the position reached at the end, StopIteration is raised.
+  # forward. When the position reached at the end, StopIteration is raised.
   #
   # === Example
   #
@@ -317,7 +321,7 @@ class Enumerator
   #   e.next_values   -> array
   #
   # Returns the next object as an array in the enumerator, and move the
-  # internal position forward.  When the position reached at the end,
+  # internal position forward. When the position reached at the end,
   # StopIteration is raised.
   #
   # This method can be used to distinguish <code>yield</code> and <code>yield
@@ -401,7 +405,7 @@ class Enumerator
   #   e.peek   -> object
   #
   # Returns the next object in the enumerator, but doesn't move the internal
-  # position forward.  If the position is already at the end, StopIteration
+  # position forward. If the position is already at the end, StopIteration
   # is raised.
   #
   # === Example
@@ -425,7 +429,7 @@ class Enumerator
   #   e.peek_values   -> array
   #
   # Returns the next object as an array, similar to Enumerator#next_values, but
-  # doesn't move the internal position forward.  If the position is already at
+  # doesn't move the internal position forward. If the position is already at
   # the end, StopIteration is raised.
   #
   # === Example
@@ -554,6 +558,46 @@ class Enumerator
       self
     end
   end
+
+  ##
+  # call-seq:
+  #    Enumerator.produce(initial = nil) { |val| } -> enumerator
+  #
+  # Creates an infinite enumerator from any block, just called over and
+  # over. Result of the previous iteration is passed to the next one.
+  # If +initial+ is provided, it is passed to the first iteration, and
+  # becomes the first element of the enumerator; if it is not provided,
+  # first iteration receives +nil+, and its result becomes first
+  # element of the iterator.
+  #
+  # Raising StopIteration from the block stops an iteration.
+  #
+  # Examples of usage:
+  #
+  #   Enumerator.produce(1, &:succ)   # => enumerator of 1, 2, 3, 4, ....
+  #
+  #   Enumerator.produce { rand(10) } # => infinite random number sequence
+  #
+  #   ancestors = Enumerator.produce(node) { |prev| node = prev.parent or raise StopIteration }
+  #   enclosing_section = ancestors.find { |n| n.type == :section }
+  def Enumerator.produce(init=NONE, &block)
+    raise ArgumentError, "no block given" if block.nil?
+    Enumerator.new do |y|
+      if init == NONE
+        val = nil
+      else
+        val = init
+        y.yield(val)
+      end
+      begin
+        while true
+          y.yield(val = block.call(val))
+        end
+      rescue StopIteration
+        # do nothing
+      end
+    end
+  end
 end
 
 module Kernel
@@ -561,15 +605,10 @@ module Kernel
   # call-seq:
   #   obj.to_enum(method = :each, *args)                 -> enum
   #   obj.enum_for(method = :each, *args)                -> enum
-  #   obj.to_enum(method = :each, *args) {|*args| block} -> enum
-  #   obj.enum_for(method = :each, *args){|*args| block} -> enum
   #
   # Creates a new Enumerator which will enumerate by calling +method+ on
   # +obj+, passing +args+ if any.
   #
-  # If a block is given, it will be used to calculate the size of
-  # the enumerator without the need to iterate it (see Enumerator#size).
-  #
   # === Examples
   #
   #   str = "xyz"
@@ -587,17 +626,14 @@ module Kernel
   # It is typical to call to_enum when defining methods for
   # a generic Enumerable, in case no block is passed.
   #
-  # Here is such an example, with parameter passing and a sizing block:
+  # Here is such an example with parameter passing:
   #
   #     module Enumerable
   #       # a generic method to repeat the values of any enumerable
   #       def repeat(n)
   #         raise ArgumentError, "#{n} is negative!" if n < 0
   #         unless block_given?
-  #           return to_enum(__method__, n) do # __method__ is :repeat here
-  #             sz = size     # Call size and multiply by n...
-  #             sz * n if sz  # but return nil if size itself is nil
-  #           end
+  #           return to_enum(__callee__, n) do # __callee__ is :repeat here
   #         end
   #         each do |*val|
   #           n.times { yield *val }
@@ -612,9 +648,6 @@ module Kernel
   #     enum.first(4) # => [1, 1, 1, 2]
   #
   def to_enum(meth=:each, *args)
-    unless self.respond_to?(meth)
-      raise ArgumentError, "undefined method #{meth}"
-    end
     Enumerator.new self, meth, *args
   end
   alias enum_for to_enum
